#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <sysexits.h>
#include <roboctl/roboctl.h>
#include <libgamepad/gamepad.h>

/* 
 *  FreeBSD:
 *  
 *  For instructions on enabling the PC joystick port, run "man joy"
 *
 *  /dev/joy returns values roughly from 0 to 1000 for x and y.
 *  Joysticks can be noisy: x and y are not steady, but spontaneously
 *  vary.  Noise on my Kraft joystick is generally only 1 or 2.  My
 *  (cheap) PC Mission varies by as much as 30 units.  Libgamepad
 *  filters out the noise by ignoring small changes.
 */

/* Robot parameters */
#define LEFT_MOTOR_PORT         2
#define RIGHT_MOTOR_PORT        0
#define IMPLEMENT_MOTOR_PORT    1

#ifndef ABS
#define ABS(x) ((x) > 0 ? (x) : -(x))
#endif

typedef struct
{
    char    *device;
    double  wheel_power;
    int     implement_power;
    int     button;
}   settings_t;

int     joy_scaled_x(int x,int max_x);
int     joy_scaled_y(int y,int max_y);
void    control_motion(rct_brick_t *brick,int x,int y,settings_t *settings);
void    control_implement_with_button(rct_brick_t *brick,int b1,int b2,settings_t *settings);
void    control_implement_with_joystick(rct_brick_t *brick,int z,int max_z,settings_t *settings);
void    swap(int *,int *);
int     process_args(int argc,char *argv[],settings_t *settings);

int     main(int argc,char *argv[])

{
    int     x, y, c;
    extern int  Debug;
    rct_brick_list_t    bricks;
    rct_brick_t         *brick;
    gamepad_t           *gp;
    settings_t          settings;
    
    Debug = 0;
    
    process_args(argc,argv,&settings);
    
    /* Open gamepad. Use O_NONBLOCK to put USB gamepads into immediate
       mode.  This allows sampling of the device state instead of 
       the device sending interrupts for every change.  Analog joysticks
       really always work this way, but the gameport_analog_read() routine
       simulates blocking I/O by looping until the device state changes
       (with usleep() to keep CPU usage down) if O_NONBLOCK is not set. */
    gp = gamepad_open(settings.device,GAMEPAD_INTERFACE_AUTO,
		    O_RDONLY|O_NONBLOCK);
    if ( gp == NULL )
    {
	fprintf(stderr,"Cannot open %s.\n",settings.device);
	return EX_UNAVAILABLE;
    }
    
    /* Open connection to NXT. */
    rct_find_bricks(&bricks,RCT_PROBE_DEV_NXT);
    brick = rct_get_brick_from_list(&bricks,0);
    if ( rct_open_brick(brick) != RCT_OK )
    {
	return EX_UNAVAILABLE;
    }
    rct_print_device_info(brick);
    
    /* Disable NXT response packets to reduce communication overhead. */
    nxt_response_off(&brick->nxt);

    c =0;
    while ( gamepad_read(gp) >= 0 )
    {
	/* Get joystick coordinates in Cartesian plane from -100 to +100 */
	x = joy_scaled_x(gamepad_x(gp),gamepad_max_x(gp));
	y = joy_scaled_y(gamepad_y(gp),gamepad_max_y(gp));
	control_motion(brick,x,y,&settings);

	if ( settings.button )
	    control_implement_with_button(brick,gamepad_button(gp,1),
		    gamepad_button(gp,2),&settings);
	else
	    control_implement_with_joystick(brick,
		    gamepad_rz(gp),gamepad_max_rz(gp),
		    &settings);

	/* Since we're using non-blocking (immediate) mode,
	   cut joystick sampling rate to about 20/sec so that 
	   NXT can keep up.  Each Bluetooth packet takes about 30ms. */
	usleep(50000);
	
	/* NXT will go to sleep unless it's running a program, so reset
	   timer occasionally. */
	if ( ++c % 1000 == 0 )
	    nxt_keep_alive(&brick->nxt);
    }
    
    gamepad_close(gp);
    rct_close_brick(brick);
    return EX_OK;
}


/*
 *  Convert raw joystick input (typically 0 for left/up and some large
 *  number for right/down) to Cartesian coordinates betweeen -100 and 100.
 */

int     joy_scaled_x(int x,int max_x)

{
    /* Scale to range of -100 (left) to +100 (right) */
    x = (x - (max_x/2)) / (max_x/200);
    if ( x < 0 )
	return MAX(x,-100);
    else
	return MIN(x,100);
}


int     joy_scaled_y(int y,int max_y)

{
    /* Scale to range of -100 (down) to +100 (up) */
    y = -((y - (max_y/2)) / (max_y/200));
    if ( y < 0 )
	return MAX(y,-100);
    else
	return MIN(y,100);
}


/****************************************************************************
 * Description: 
 *  Convert joystick x and y coordinates (-100 to 100) to left and
 *  right motor speed (also -100 to 100) using skid steering.
 *  It's a little tricky to control both motors with one joystick.
 *  Below is a map of the joystick Cartesian plane, showing the desired
 *  left and right motor speeds (l,r) at each point (x,y).  This map was
 *  designed by first defining vehicle behavior at x,y = 0,0, and at the
 *  extremes, where x or y is -100 or 100.  Points between are linearly
 *  interpolated.
 *
 *  For example:
 *      x,y = 0,0:      Vehicle is motionless.
 *      x,y = 0,100:    Full speed straight ahead.
 *      x,y = 100,0:    Full speed rotating clockwise about center.
 *                      (Left side sull speed ahead, right full speed reverse)
 *      x,y = 100,100:  Full speed rotating about right wheel/track.
 *                      (Left side full speed, right side stopped.)
 *      all x>0:        Vehicle is rotating clockwise.
 *      all x<0:        Vehicle is rotating counter-clockwise.
 *  
 *      0,100       50,100      100,100     100,50      100,0
 *
 *      -50,100     0,75        50,50       75,0        100,-50
 *
 *      -100,100    -50,50      0,0         50,-50      100,-100
 *
 *      -100,50     -75,0       -50,-50     0,-75       50,-100
 *
 *      -100,0      -100,-50    -100,-100   -50,-100    0,-100
 *  
 *  The formula for a given quandrant is simple.  A general formula for
 *  the entire plane would be ugly.  ( Observe how l varies across
 *  the top of the plane (where y=100), increasing from 0 to 100 over x from
 *  -100 to 0, and then becoming a constant 100 for x from 0 to 100. )
 *  
 *  The values are highly symmetric between various quandrants.  L and r
 *  are basically swapped going from any x to -x, and swapped and negated
 *  for any y and -y.  Hence, we simply compute (l,r) for the first quadrant
 *  by using abs(x) and abs(y), and then swap and/or negate as needed
 *  for the other 3 quadrants.
 ***************************************************************************/

void    control_motion(rct_brick_t *brick,int x,int y,settings_t *settings)

{
    int     left_speed,
	    right_speed,
	    abs_x = ABS(x),
	    abs_y = ABS(y);

    /*
     * Compute left and right motor speeds using simple formula for
     * quadrant 1.
     */
    left_speed = abs_y + abs_x * ( 100 - abs_y ) / 100;
    right_speed = abs_y - abs_x;

    /* 
     * Vehicle doesn't move until power reaches 60 or higher.
     * Warp power function to increase slope of power function 
     * for small x & y where vehicle isn't moving, and decrease
     * it for high power.
     * Ex. power = sqrt(power) * 10 produces power in the range
     * of 0 to 100, but with a steadily decreasing slope.
     * This will make the joystick less sensitive over the range
     * where vehicle speed actually varies.  Some experimentation
     * revealed that higher roots make driving progressively easier,
     * with the 4th root (y=x^0.25) working best.  The 5th
     * root makes the robot move with a joystick position of 1,
     * which means the robot will not stop unless the joystick
     * is perfectly centered.
     */
    
    /* Warp by taking the 4th root of the motor power, and normalize back
       to a range of 0 to 100 by dividing by 100^(1/4) */
    left_speed = settings->wheel_power * pow((double)left_speed,0.25) / 3.16;
    if ( right_speed >= 0 )
	right_speed = settings->wheel_power *
			pow((double)right_speed,0.25) / 3.16;
    else
	right_speed = -settings->wheel_power *
			pow((double)-right_speed,0.25) / 3.16;
    
    if ( y >= 0 )   /* Quadrants 1 and 2 */
    {
	/* 
	 *  Left speed and right speed are simply swapped when going from
	 *  x,y to -x,y.  ( Quandrant 1 to 2 )
	 */
	if ( x < 0 )    /* Swap for quadrant 2 */
	    swap(&left_speed,&right_speed);
    }
    else    /* Quadrants 3 and 4 */
    {
	/*
	 *  Left speed and right speed are swapped AND negated when going
	 *  from x,y to x,-y.  ( Quadrant 1 to 4 )
	 *
	 *  Going from x,y to -x,-y (quadrant 1 to 3), motor speeds are
	 *  only negated.
	 */
	if ( x >= 0 )   /* Swap for quadrant 4 only */
	    swap(&left_speed,&right_speed);

	/* Negate for quadrants 3 and 4 */
	left_speed = -left_speed;
	right_speed = -right_speed;
    }
    
    printf("Joystick: %4d %4d   Power: %4d %4d  ",
	    x,y,left_speed,right_speed);
    fflush(stdout);
    
    nxt_set_output_state(&(brick->nxt),LEFT_MOTOR_PORT,left_speed,
	NXT_MODE_MOTORON,NXT_REGULATION_MODE_IDLE,
	0,NXT_RUN_STATE_RUNNING,0);
    nxt_set_output_state(&(brick->nxt),RIGHT_MOTOR_PORT,right_speed,
	NXT_MODE_MOTORON,NXT_REGULATION_MODE_IDLE,
	0,NXT_RUN_STATE_RUNNING,0);
}


/**
 *  Control a simple motor-driven implement using two joystick/gamepad
 *  buttons.  Full power one way if button1 is pressed, and full power the
 *  other way if button2 is pressed.
 */

void    control_implement_with_button(rct_brick_t *brick,int b1,int b2,settings_t *settings)

{
    static int  old_implement_power = 0,
		implement_power;
    
    if ( b2 == 1 )
	implement_power = settings->implement_power;
    else if ( b1 == 1 )
	implement_power = -settings->implement_power;
    else
	implement_power=0;
    
    if ( implement_power != old_implement_power )
    {
	nxt_set_output_state(&(brick->nxt),IMPLEMENT_MOTOR_PORT,
	    implement_power,
	    NXT_MODE_MOTORON,NXT_REGULATION_MODE_IDLE,
	    0,NXT_RUN_STATE_RUNNING,0);
    }
    printf("Implement power: %4d\r",implement_power);
    old_implement_power = implement_power;
}


/**
 *  Control a simple motor-driven implement by mapping joystick Y 
 *  position information to motor power.
 */

void    control_implement_with_joystick(rct_brick_t *brick,int z,int max_z,
			    settings_t *settings)

{
    static int  old_implement_power = 0,
		implement_power;
    
    
    /* Get percent power between 0 and 100 */
    implement_power = 100 * (2 * z - max_z) / max_z;
    
    /* Warp by taking the 4th root of the motor power, and normalize back
       to a range of 0 to 100 by dividing by 100^(1/4) */
    if ( implement_power >= 0 )
	implement_power = settings->implement_power * 
				pow((double)implement_power,0.25) / 3.16;
    else
	implement_power = -settings->implement_power *
				pow((double)-implement_power,0.25) / 3.16;
    
    if ( implement_power != old_implement_power )
    {
	nxt_set_output_state(&(brick->nxt),IMPLEMENT_MOTOR_PORT,
	    implement_power,
	    NXT_MODE_MOTORON,NXT_REGULATION_MODE_IDLE,
	    0,NXT_RUN_STATE_RUNNING,0);
    }
    printf("Implement power: %4d\r",implement_power);
    old_implement_power = implement_power;
}


void    swap(int *a,int *b)

{
    int     temp = *a;
    
    *a = *b;
    *b = temp;
}


int     process_args(int argc,char *argv[],settings_t *settings)

{
    char    *end;
    
    switch(argc)
    {
	case    5:
	    settings->device=argv[1];
	    settings->wheel_power = strtol(argv[2],&end,10);
	    if ( end-argv[2] != strlen(argv[2]) )
	    {
		fprintf(stderr,"%s is not a valid power value.  Must be from 0 to 100.\n",argv[2]);
		exit(EX_USAGE);
	    }
	    settings->implement_power = strtol(argv[3],&end,10);
	    if ( end-argv[3] != strlen(argv[3]) )
	    {
		fprintf(stderr,"%s is not a valid power value.  Must be from 0 to 100.\n",argv[3]);
		exit(EX_USAGE);
	    }
	    if ( strcmp(argv[4],"button") == 0 )
		settings->button = 1;
	    else if ( strcmp(argv[4],"joystick") == 0 )
		settings->button = 0;
	    else
		fputs("Last argument must be 'button' or 'joystick'\n",stderr);
	    break;
	default:
	    printf("\nUsage:\n\n%s <device> <max wheel power> <max implement power> <implement control>\n",argv[0]);
	    puts("\nExamples:");
	    printf("    %s /dev/joy0 85 63 button\n",argv[0]);
	    printf("    %s /dev/uhid0 90 68 joystick\n",argv[0]);
	    exit(EX_USAGE);
    }
    return 0;
}

